# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import fnmatch
import glob
import logging as log
import os
import os.path as osp
from typing import List, Optional

from datumaro.components.annotation import AnnotationType, Label, LabelCategories
from datumaro.components.dataset import DatasetItem
from datumaro.components.dataset_base import DatasetBase
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.image import find_images


class MarsPath:
    SUBSET_DIR_PATTERN = "bbox_*"
    IMAGE_DIR_PATTERNS = ["[0-9]" * 4, "00-1"]
    IMAGE_NAME_POSTFIX = "C[0-9]" + "T" + "[0-9]" * 4 + "F" + "[0-9]" * 3 + ".jpg"


class MarsBase(DatasetBase):
    def __init__(self, path: str, *, ctx: Optional[ImportContext] = None):
        assert osp.isdir(path), path
        super().__init__(ctx=ctx)

        self._dataset_dir = path
        self._subsets = {
            subset_dir.split("_", maxsplit=1)[1]: osp.join(path, subset_dir)
            for subset_dir in os.listdir(path)
            if (
                osp.isdir(osp.join(path, subset_dir))
                and fnmatch.fnmatch(subset_dir, MarsPath.SUBSET_DIR_PATTERN)
            )
        }

        self._categories = self._load_categories()
        self._items = []
        for subset, subset_path in self._subsets.items():
            self._items.extend(self._load_items(subset, subset_path))

    def __iter__(self):
        yield from self._items

    def categories(self):
        return self._categories

    def _load_categories(self):
        dirs = sorted(
            [
                dir_name
                for subset_path in self._subsets.values()
                for dir_name in os.listdir(subset_path)
                if (
                    osp.isdir(osp.join(self._dataset_dir, subset_path, dir_name))
                    and any(
                        fnmatch.fnmatch(dir_name, image_dir)
                        for image_dir in MarsPath.IMAGE_DIR_PATTERNS
                    )
                )
            ]
        )
        return {AnnotationType.label: LabelCategories.from_iterable(dirs)}

    def _load_items(self, subset, path):
        items = []
        for label_cat in self._categories[AnnotationType.label]:
            label = label_cat.name
            label_id = self._categories[AnnotationType.label].find(label)[0]
            for image_path in find_images(osp.join(path, label)):
                image_name = osp.basename(image_path)
                item_id = osp.splitext(image_name)[0]
                pedestrian_id = image_name[0:4]

                if not fnmatch.fnmatch(image_name, label + MarsPath.IMAGE_NAME_POSTFIX):
                    items.append(DatasetItem(id=item_id, image=image_path))
                    continue

                if pedestrian_id != label:
                    log.warning(
                        f"The image {image_path} will be skip because"
                        "pedestrian id for it does not match with"
                        f"the directory name: {label}"
                    )
                    continue

                items.append(
                    DatasetItem(
                        id=item_id,
                        media=Image.from_file(path=image_path),
                        subset=subset,
                        annotations=[Label(label=label_id)],
                        attributes={
                            "person_id": pedestrian_id,
                            "camera_id": int(image_name[5]),
                            "track_id": int(image_name[7:11]),
                            "frame_id": int(image_name[12:15]),
                        },
                    )
                )
                self._ann_types.add(AnnotationType.label)

        return items


class MarsImporter(Importer):
    @classmethod
    def detect(cls, context: FormatDetectionContext):
        with context.require_any():
            for image_dir in MarsPath.IMAGE_DIR_PATTERNS:
                with context.alternative():
                    context.require_file(
                        "/".join(
                            [
                                MarsPath.SUBSET_DIR_PATTERN,
                                image_dir,
                                image_dir + MarsPath.IMAGE_NAME_POSTFIX,
                            ]
                        )
                    )

    @classmethod
    def find_sources(cls, path):
        patterns = [
            "/".join((path, subset_dir, image_dir, image_dir + MarsPath.IMAGE_NAME_POSTFIX))
            for image_dir in MarsPath.IMAGE_DIR_PATTERNS
            for subset_dir in os.listdir(path)
            if (
                osp.isdir(osp.join(path, subset_dir))
                and fnmatch.fnmatch(subset_dir, MarsPath.SUBSET_DIR_PATTERN)
            )
        ]

        for pattern in patterns:
            try:
                next(glob.iglob(pattern))
                return [{"url": path, "format": "mars"}]
            except StopIteration:
                continue

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [osp.splitext(MarsPath.IMAGE_NAME_POSTFIX)[1]]



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from typing import List

from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer

from .format import SynthiaAlPath, SynthiaFormatType, SynthiaRandPath, SynthiaSfPath


class _SynthiaImporter(Importer):
    FORMAT = None
    META_FOLDERS = []

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        for folder in cls.META_FOLDERS:
            if not osp.isdir(osp.join(context.root_path, folder)):
                context.fail("Any Synthia format is not detected.")

        return FormatDetectionConfidence.MEDIUM

    @classmethod
    def find_sources(cls, path):
        return [{"url": path, "format": cls.FORMAT}]

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [".png"]


class SynthiaRandImporter(_SynthiaImporter):
    FORMAT = SynthiaFormatType.synthia_rand.name
    META_FOLDERS = SynthiaRandPath.meta_folders()


class SynthiaSfImporter(_SynthiaImporter):
    FORMAT = SynthiaFormatType.synthia_sf.name
    META_FOLDERS = SynthiaSfPath.meta_folders()


class SynthiaAlImporter(_SynthiaImporter):
    FORMAT = SynthiaFormatType.synthia_al.name
    META_FOLDERS = SynthiaAlPath.meta_folders()



# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import gzip
import os
import os.path as osp
from typing import List, Optional

import numpy as np

from datumaro.components.annotation import AnnotationType, Label, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file


class MnistPath:
    TEST_LABELS_FILE = "t10k-labels-idx1-ubyte.gz"
    TEST_IMAGES_FILE = "t10k-images-idx3-ubyte.gz"
    LABELS_FILE = "-labels-idx1-ubyte.gz"
    IMAGES_FILE = "-images-idx3-ubyte.gz"
    IMAGE_SIZE = 28
    NONE_LABEL = 255


class MnistBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find annotations file", path)
        self._dataset_dir = osp.dirname(path)

        if not subset:
            file_name = osp.splitext(osp.basename(path))[0]
            if file_name.startswith("t10k"):
                subset = "test"
            else:
                subset = file_name.split("-", maxsplit=1)[0]

        super().__init__(subset=subset, ctx=ctx)

        self._categories = self._load_categories()
        self._items = list(self._load_items(path).values())

    def _load_categories(self):
        if has_meta_file(self._dataset_dir):
            return {
                AnnotationType.label: LabelCategories.from_iterable(
                    parse_meta_file(self._dataset_dir).keys()
                )
            }

        label_cat = LabelCategories()

        labels_file = osp.join(self._dataset_dir, "labels.txt")
        if osp.isfile(labels_file):
            with open(labels_file, encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue
                    label_cat.add(line)
        else:
            for i in range(10):
                label_cat.add(str(i))

        return {AnnotationType.label: label_cat}

    def _load_items(self, path):
        items = {}
        with gzip.open(path, "rb") as lbpath:
            labels = np.frombuffer(lbpath.read(), dtype=np.uint8, offset=8)

        meta = []
        metafile = osp.join(self._dataset_dir, self._subset + "-meta.gz")
        if osp.isfile(metafile):
            with gzip.open(metafile, "rb") as f:
                meta = np.frombuffer(f.read(), dtype="<U32")
            meta = meta.reshape(len(labels), int(len(meta) / len(labels)))

        # support for single-channel image only
        images = None
        images_file = osp.join(
            self._dataset_dir, osp.basename(path).replace("labels-idx1", "images-idx3")
        )
        if osp.isfile(images_file):
            with gzip.open(images_file, "rb") as imgpath:
                images = np.frombuffer(imgpath.read(), dtype=np.uint8, offset=16)
                if len(meta) == 0 or len(meta[0]) < 2:
                    images = images.reshape(len(labels), MnistPath.IMAGE_SIZE, MnistPath.IMAGE_SIZE)

        pix_num = 0
        for i, annotation in enumerate(labels):
            annotations = []
            label = annotation
            if label != MnistPath.NONE_LABEL:
                annotations.append(Label(label))
                self._ann_types.add(AnnotationType.label)

            image = None
            if images is not None:
                if 0 < len(meta) and 1 < len(meta[i]):
                    h, w = int(meta[i][-2]), int(meta[i][-1])
                    image = images[pix_num : pix_num + h * w].reshape(h, w)
                    pix_num += h * w
                else:
                    image = images[i].reshape(MnistPath.IMAGE_SIZE, MnistPath.IMAGE_SIZE)

            if image is not None:
                image = Image.from_numpy(data=image)

            if 0 < len(meta) and (len(meta[i]) == 1 or len(meta[i]) == 3):
                i = meta[i][0]

            items[i] = DatasetItem(id=i, subset=self._subset, media=image, annotations=annotations)

        return items


class MnistImporter(Importer):
    _FORMAT_EXT = ".gz"

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(
            path,
            cls._FORMAT_EXT,
            "mnist",
            file_filter=lambda p: 1 < len(osp.basename(p).split("-"))
            and osp.basename(p).split("-")[1] == "labels",
        )

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._FORMAT_EXT]


class MnistExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".png"

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        os.makedirs(self._save_dir, exist_ok=True)
        if self._save_dataset_meta:
            self._save_meta_file(self._save_dir)

        for subset_name, subset in self._extractor.subsets().items():
            labels = []
            images = np.array([])
            item_ids = {}
            image_sizes = {}
            for item in subset:
                anns = [a.label for a in item.annotations if a.type == AnnotationType.label]
                label = 255
                if anns:
                    label = anns[0]
                labels.append(label)

                if item.id != str(len(labels) - 1):
                    item_ids[len(labels) - 1] = item.id

                if item.media and self._save_media:
                    image = item.media
                    if not image.has_data:
                        image_sizes[len(images) - 1] = [0, 0]
                    else:
                        image = image.data
                        if (
                            image.shape[0] != MnistPath.IMAGE_SIZE
                            or image.shape[1] != MnistPath.IMAGE_SIZE
                        ):
                            image_sizes[len(labels) - 1] = [image.shape[0], image.shape[1]]
                        images = np.append(images, image.reshape(-1).astype(np.uint8))

            if subset_name == "test":
                labels_file = osp.join(self._save_dir, MnistPath.TEST_LABELS_FILE)
            else:
                labels_file = osp.join(self._save_dir, subset_name + MnistPath.LABELS_FILE)
            self.save_annotations(labels_file, labels)

            if 0 < len(images):
                if subset_name == "test":
                    images_file = osp.join(self._save_dir, MnistPath.TEST_IMAGES_FILE)
                else:
                    images_file = osp.join(self._save_dir, subset_name + MnistPath.IMAGES_FILE)
                self.save_images(images_file, images)

            # it is't in the original format,
            # this is for storng other names and sizes of images
            if len(item_ids) or len(image_sizes):
                meta = []
                if len(item_ids) and len(image_sizes):
                    # other names and sizes of images
                    size = [MnistPath.IMAGE_SIZE, MnistPath.IMAGE_SIZE]
                    for i in range(len(labels)):
                        w, h = image_sizes.get(i, size)
                        meta.append([item_ids.get(i, i), w, h])

                elif len(item_ids):
                    # other names of images
                    for i in range(len(labels)):
                        meta.append([item_ids.get(i, i)])

                elif len(image_sizes):
                    # other sizes of images
                    size = [MnistPath.IMAGE_SIZE, MnistPath.IMAGE_SIZE]
                    for i in range(len(labels)):
                        meta.append(image_sizes.get(i, size))

                metafile = osp.join(self._save_dir, subset_name + "-meta.gz")
                with gzip.open(metafile, "wb") as f:
                    f.write(np.array(meta, dtype="<U32").tobytes())

        self.save_labels()

    def save_annotations(self, path, data):
        with gzip.open(path, "wb") as f:
            # magic number = 0x0801 (2049, hexadecimal representation)
            # this is used to verify the file with MNIST mark data
            f.write(np.array([0x0801, len(data)], dtype=">i4").tobytes())
            f.write(np.array(data, dtype="uint8").tobytes())

    def save_images(self, path, data):
        with gzip.open(path, "wb") as f:
            # magic number = 0x0803 (2051, hexadecimal representation),
            # this is used to verify the file with MNIST image data
            f.write(
                np.array(
                    [0x0803, len(data), MnistPath.IMAGE_SIZE, MnistPath.IMAGE_SIZE], dtype=">i4"
                ).tobytes()
            )
            f.write(np.array(data, dtype="uint8").tobytes())

    def save_labels(self):
        labels_file = osp.join(self._save_dir, "labels.txt")
        with open(labels_file, "w", encoding="utf-8") as f:
            f.writelines(
                l.name + "\n"
                for l in self._extractor.categories().get(AnnotationType.label, LabelCategories())
            )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from copy import copy
from enum import Enum, auto
from typing import Any, Iterator, Optional, Set, Tuple, Type, Union

from datumaro.components.annotation import AnnotationType
from datumaro.components.dataset_base import CategoriesInfo, DatasetInfo, DatasetItem, IDataset
from datumaro.components.media import MediaElement
from datumaro.util.definitions import DEFAULT_SUBSET_NAME

__all__ = ["ItemStatus", "DatasetItemStorage", "DatasetItemStorageDatasetView"]


class ItemStatus(Enum):
    added = auto()
    modified = auto()
    removed = auto()


class DatasetItemStorage:
    def __init__(self):
        self.data = {}  # { subset_name: { id: DatasetItem } }
        self._traversal_order = {}  # maintain the order of elements
        self._order = []  # allow indexing

    def __iter__(self) -> Iterator[DatasetItem]:
        for item in self._traversal_order.values():
            yield item

    def __len__(self) -> int:
        return len(self._traversal_order)

    def is_empty(self) -> bool:
        # Subsets might contain removed items, so this may differ from __len__
        return all(len(s) == 0 for s in self.data.values())

    def put(self, item: DatasetItem) -> bool:
        subset = self.data.setdefault(item.subset, {})
        is_new = subset.get(item.id) is None
        self._traversal_order[(item.id, item.subset)] = item
        if is_new:
            self._order.append((item.id, item.subset))
        subset[item.id] = item
        return is_new

    def get(
        self, id: Union[str, DatasetItem], subset: Optional[str] = None, dummy: Any = None
    ) -> Optional[DatasetItem]:
        if isinstance(id, DatasetItem):
            id, subset = id.id, id.subset
        else:
            id = str(id)
            subset = subset or DEFAULT_SUBSET_NAME

        return self.data.get(subset, {}).get(id, dummy)

    def remove(self, id: Union[str, DatasetItem], subset: Optional[str] = None) -> bool:
        if isinstance(id, DatasetItem):
            id, subset = id.id, id.subset
        else:
            id = str(id)
            subset = subset or DEFAULT_SUBSET_NAME

        subset_data = self.data.setdefault(subset, {})
        is_removed = subset_data.get(id) is not None
        subset_data[id] = None
        if is_removed:
            # TODO : investigate why "del subset_data[id]" cannot replace "subset_data[id] = None".
            self._traversal_order.pop((id, subset))
            self._order.remove((id, subset))
        return is_removed

    def __contains__(self, x: Union[DatasetItem, Tuple[str, str]]) -> bool:
        if not isinstance(x, tuple):
            x = [x]
        dummy = 0
        return self.get(*x, dummy=dummy) is not dummy

    def get_subset(self, name):
        return self.data.get(name, {})

    def subsets(self):
        return self.data

    def get_annotated_items(self):
        return sum(bool(s.annotations) for s in self._traversal_order.values())

    def get_datasetitem_by_path(self, path):
        for s in self._traversal_order.values():
            if getattr(s.media, "path", None) == path:
                return s

    def get_annotations(self):
        annotations_by_type = {t.name: {"count": 0} for t in AnnotationType}
        for item in self._traversal_order.values():
            for ann in item.annotations:
                annotations_by_type[ann.type.name]["count"] += 1
        return sum(t["count"] for t in annotations_by_type.values())

    def __copy__(self):
        copied = DatasetItemStorage()
        copied._traversal_order = copy(self._traversal_order)
        copied._order = copy(self._order)
        copied.data = copy(self.data)
        return copied

    def __getitem__(self, idx: int) -> DatasetItem:
        _id, subset = self._order[idx]
        item = self.data[subset][_id]
        return item


class DatasetItemStorageDatasetView(IDataset):
    class Subset(IDataset):
        def __init__(self, parent: "DatasetItemStorageDatasetView", name: str):
            super().__init__()
            self.parent = parent
            self.name = name
            self._length = None

        @property
        def _data(self):
            return self.parent._get_subset_data(self.name)

        def __iter__(self):
            for item in self._data.values():
                if item:
                    yield item

        def __len__(self):
            if self._length is not None:
                return self._length

            self._length = 0
            for item in self._data.values():
                if item is not None:
                    self._length += 1
            return self._length

        def put(self, item):
            self._length = None
            return self._data.put(item)

        def get(self, id, subset=None):
            assert (subset or DEFAULT_SUBSET_NAME) == (self.name or DEFAULT_SUBSET_NAME)
            return self._data.get(id, subset)

        def remove(self, id, subset=None):
            assert (subset or DEFAULT_SUBSET_NAME) == (self.name or DEFAULT_SUBSET_NAME)
            self._length = None
            return self._data.remove(id, subset)

        def get_subset(self, name):
            assert (name or DEFAULT_SUBSET_NAME) == (self.name or DEFAULT_SUBSET_NAME)
            return self

        def subsets(self):
            return {self.name or DEFAULT_SUBSET_NAME: self}

        def infos(self):
            return self.parent.infos()

        def categories(self):
            return self.parent.categories()

        def media_type(self):
            return self.parent.media_type()

        def ann_types(self):
            return self.parent.ann_types()

    def __init__(
        self,
        parent: DatasetItemStorage,
        infos: DatasetInfo,
        categories: CategoriesInfo,
        media_type: Optional[Type[MediaElement]],
        ann_types: Optional[Set[AnnotationType]],
    ):
        self._parent = parent
        self._infos = infos
        self._categories = categories
        self._media_type = media_type
        self._ann_types = ann_types

    def __iter__(self):
        yield from self._parent

    def __len__(self):
        return len(self._parent)

    def infos(self):
        return self._infos

    def categories(self):
        return self._categories

    def get_subset(self, name):
        return self.Subset(self, name)

    def _get_subset_data(self, name):
        return self._parent.get_subset(name)

    def subsets(self):
        return {k: self.get_subset(k) for k in self._parent.subsets()}

    def get(self, id, subset=None):
        return self._parent.get(id, subset=subset)

    def media_type(self):
        return self._media_type

    def ann_types(self):
        return self._ann_types



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import contextlib
import logging as log
import os
import os.path as osp
import sys
from typing import Dict, Tuple

from datumaro.components.dataset_base import IDataset
from datumaro.components.environment import DEFAULT_ENVIRONMENT
from datumaro.components.extractor_tfds import (
    AVAILABLE_TFDS_DATASETS,
    TFDS_EXTRACTOR_AVAILABLE,
    TfdsDatasetRemoteMetadata,
)
from datumaro.util import dump_json
from datumaro.util.os_util import make_file_name

from ...util.errors import CliException
from ...util.project import generate_next_file_name
from .downloader import IDatasetDownloader


class TfdsDatasetDownloader(IDatasetDownloader):
    @classmethod
    def get_extractor(cls, dataset_id: str) -> Tuple[str, IDataset]:
        if dataset_id.startswith("tfds:"):
            if TFDS_EXTRACTOR_AVAILABLE:
                tfds_ds_name = dataset_id[5:]
                tfds_ds = AVAILABLE_TFDS_DATASETS.get(tfds_ds_name)
                if tfds_ds:
                    default_output_format = tfds_ds.metadata.default_output_format
                    extractor_factory = tfds_ds.make_extractor
                    return default_output_format, extractor_factory
                else:
                    raise CliException(f"Unsupported TFDS dataset '{tfds_ds_name}'")
            else:
                raise CliException(
                    "TFDS datasets are not available, because TFDS and/or"
                    "TensorFlow are not installed.\n"
                    "You can install them with: pip install datumaro[tf,tfds]"
                )
        else:
            raise CliException(f"Unknown dataset ID TFDS dataset '{tfds_ds_name}'")

    @staticmethod
    def _describe_txt(dataset_metas: Dict[str, TfdsDatasetRemoteMetadata], report_file=None):
        with open(report_file, "w") if report_file else contextlib.nullcontext() as report_file:
            if dataset_metas:
                print("Available datasets:", file=report_file)
                for name, meta in sorted(dataset_metas.items()):
                    print(
                        f"""
{name} ({meta.human_name}):
  default output format: {meta.default_output_format}
  description:""",
                        file=report_file,
                    )
                    for line in meta.description.rstrip("\n").split("\n"):
                        print(f"    {line}", file=report_file)
                    print(
                        f"""  download size: {meta.download_size} bytes
  home URL: {meta.home_url or 'N/A'}
  number of classes: {meta.num_classes}
  subsets:""",
                        file=report_file,
                    )
                    for subset_name, subset_meta in sorted(meta.subsets.items()):
                        print(f"    {subset_name}: {subset_meta.num_items} items", file=report_file)
                    print(f"  version: {meta.version}" "", file=report_file)
            else:
                print(
                    """No datasets available.

"You can enable TFDS datasets by installing TensorFlow and TensorFlow Datasets:
    pip install datumaro[tf,tfds]""",
                    file=report_file,
                )

    @staticmethod
    def _describe_json(dataset_metas, report_file):
        def meta_to_raw(meta: TfdsDatasetRemoteMetadata):
            raw = {}

            # We omit the media type from the output, because there is currently no mechanism
            # for mapping media types to strings. The media type could be useful information
            # for users, though, so we might want to implement such a mechanism eventually.

            for attribute in (
                "default_output_format",
                "description",
                "download_size",
                "home_url",
                "human_name",
                "num_classes",
                "version",
            ):
                raw[attribute] = getattr(meta, attribute)

            raw["subsets"] = {
                name: {"num_items": subset.num_items} for name, subset in meta.subsets.items()
            }

            return raw

        with (
            open(report_file, "w") if report_file else contextlib.nullcontext(sys.stdout)
        ) as report_file:
            report_file.write(
                dump_json(
                    {name: meta_to_raw(meta) for name, meta in dataset_metas.items()},
                    indent=True,
                    append_newline=True,
                ).decode()
            )

    @classmethod
    def describe(cls, report_format, report_file=None):
        dataset_metas: Dict[str, TfdsDatasetRemoteMetadata] = {}

        if TFDS_EXTRACTOR_AVAILABLE:
            for dataset_name, dataset in AVAILABLE_TFDS_DATASETS.items():
                dataset_metas[f"tfds:{dataset_name}"] = dataset.query_remote_metadata()

        if report_format == "text":
            cls._describe_txt(dataset_metas, report_file)

        elif report_format == "json":
            cls._describe_json(dataset_metas, report_file)

    @classmethod
    def describe_command_description(_):
        return """More detailed
        information can be found in the TFDS Catalog:
        <https://www.tensorflow.org/datasets/catalog/overview>."""

    @classmethod
    def get_command_description(cls):
        builtin_writers = sorted(DEFAULT_ENVIRONMENT.exporters)
        if TFDS_EXTRACTOR_AVAILABLE:
            available_datasets = ", ".join(f"tfds:{name}" for name in AVAILABLE_TFDS_DATASETS)
        else:
            available_datasets = "N/A (TensorFlow and/or TensorFlow Datasets are not installed)"
            return f"""
Supported datasets: {available_datasets}|n
|n
Supported output formats: {", ".join(builtin_writers)}|n
|n
Examples:|n
- Download the MNIST dataset:|n
|s|s%(prog)s -i tfds:mnist -- --save-media|n
|n
- Download the VOC 2012 dataset, saving only the annotations in the COCO
format into a specific directory:|n
|s|s%(prog)s -i tfds:voc/2012 -f coco -o path/I/like/
"""

    @classmethod
    def download(cls, dataset_id, dst_dir, overwrite, output_format, subset, extra_args):
        env = DEFAULT_ENVIRONMENT
        default_output_format, extractor_factory = cls.get_extractor(dataset_id)
        output_format = output_format or default_output_format

        try:
            exporter = env.exporters[output_format]
        except KeyError:
            raise CliException(f"Exporter for format '{output_format}' is not found")
        extra_args = exporter.parse_cmdline(extra_args)

        if dst_dir:
            if not overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
                raise CliException(
                    f"Directory '{dst_dir}' already exists (pass --overwrite to overwrite)"
                )
        else:
            dst_dir = generate_next_file_name(
                f"{make_file_name(dataset_id)}-{make_file_name(output_format)}"
            )
        dst_dir = osp.abspath(dst_dir)

        log.info("Downloading the dataset")
        extractor = extractor_factory()

        if subset:
            try:
                extractor = extractor.subsets()[subset]
            except KeyError:
                raise CliException(f"Subset '{subset}' is not present in the dataset")

        log.info("Exporting the dataset")
        exporter.convert(extractor, dst_dir, default_image_ext=".png", **extra_args)

        log.info(f"Dataset exported to '{dst_dir}' as '{output_format}'")



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.components.environment import DEFAULT_ENVIRONMENT
from datumaro.components.errors import ProjectNotFoundError
from datumaro.components.project import ProjectBuildTargets
from datumaro.util.os_util import make_file_name
from datumaro.util.scope import scope_add, scoped

from ....util import MultilineFormatter
from ....util.errors import CliException
from ....util.project import FilterModes, generate_next_file_name, load_project

__all__ = [
    "build_parser",
    "get_sensitive_args",
]


def build_parser(parser_ctor=argparse.ArgumentParser):
    builtins = sorted(DEFAULT_ENVIRONMENT.exporters)

    parser = parser_ctor(
        help="Export project",
        description="""
        Exports a project in some format.|n
        |n
        Each dataset format has its own export
        options, which are passed after the '--' separator (see examples),
        pass '-- -h' for more info. If not stated otherwise, by default
        only annotations are exported, to include images pass
        '--save-media' parameter.|n
        |n
        A filter can be passed, check the 'filter' command description for
        more info.|n
        |n
        Formats:|n
        Datasets come in a wide variety of formats. Each dataset
        format defines its own data structure and rules on how to
        interpret the data. Check the user manual for the list of
        supported formats, examples and documentation.
        |n
        The list of supported formats can be extended by plugins.
        Check the "plugins" section of the developer guide for information
        about plugin implementation.|n
        |n
        List of builtin dataset formats: {}|n
        |n
        The command can only be applied to a project build target, a stage
        or the combined 'project' target, in which case all the targets will
        be affected.
        |n
        Examples:|n
        - Export project as a VOC-like dataset, include images:|n
        |s|s%(prog)s -f voc -- --save-media|n
        |n
        - Export project as a COCO-like dataset in other directory:|n
        |s|s%(prog)s -f coco -o path/I/like/
        """.format(
            ", ".join(builtins)
        ),
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "_positionals", nargs=argparse.REMAINDER, help=argparse.SUPPRESS
    )  # workaround for -- eaten by positionals
    parser.add_argument(
        "target",
        nargs="?",
        default="project",
        help="A project target to be exported (default: %(default)s)",
    )
    parser.add_argument("-e", "--filter", help="XML XPath filter expression for dataset items")
    parser.add_argument(
        "--filter-mode",
        default=FilterModes.i.name,
        type=FilterModes.parse,
        help="Filter mode (options: %s; default: %s)"
        % (", ".join(FilterModes.list_options()), "%(default)s"),
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="Directory to save output (default: a subdir in the current one)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.add_argument("-f", "--format", required=True, help="Output format")
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for exporter (pass '-- -h' for help). "
        "Must be specified after the main command arguments and after "
        "the '--' separator",
    )
    parser.set_defaults(command=export_command)

    return parser


def get_sensitive_args():
    return {
        export_command: ["dst_dir", "project_dir", "name", "extra_args", "target", "filter"],
    }


@scoped
def export_command(args):
    has_sep = "--" in args._positionals
    if has_sep:
        pos = args._positionals.index("--")
        if 1 < pos:
            raise argparse.ArgumentError(None, message="Expected no more than 1 target argument")
    else:
        pos = 1
    args.target = (args._positionals[:pos] or [ProjectBuildTargets.MAIN_TARGET])[0]
    args.extra_args = args._positionals[pos + has_sep :]

    show_plugin_help = "-h" in args.extra_args or "--help" in args.extra_args

    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if not show_plugin_help:
            raise

    if project is not None:
        env = project.env
    else:
        env = DEFAULT_ENVIRONMENT

    try:
        exporter = env.exporters[args.format]
    except KeyError:
        raise CliException("Exporter for format '%s' is not found" % args.format)

    extra_args = exporter.parse_cmdline(args.extra_args)

    dst_dir = args.dst_dir
    if dst_dir:
        if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
            raise CliException(
                "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
            )
    else:
        dst_dir = generate_next_file_name("export-%s" % make_file_name(args.format))
    dst_dir = osp.abspath(dst_dir)

    if args.filter:
        filter_args = FilterModes.make_filter_args(args.filter_mode)
        filter_expr = args.filter

    log.info("Loading the project...")

    dataset = project.working_tree.make_dataset(args.target)
    if args.filter:
        dataset.filter(filter_expr, **filter_args)

    log.info("Exporting...")

    dataset.export(save_dir=dst_dir, format=exporter, **extra_args)

    log.info("Results have been saved to '%s'" % dst_dir)

    return 0



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
import struct
from io import BufferedReader
from multiprocessing.pool import AsyncResult, Pool
from typing import Any, Dict, List, Optional

from datumaro.components.crypter import NULL_CRYPTER, Crypter
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import DatasetImportError
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image, MediaElement, MediaType, PointCloud, Video, VideoFrame
from datumaro.plugins.data_formats.datumaro_binary.format import DatumaroBinaryPath
from datumaro.plugins.data_formats.datumaro_binary.mapper import DictMapper
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import IntListMapper
from datumaro.plugins.data_formats.datumaro_binary.mapper.dataset_item import DatasetItemMapper

from ..datumaro.base import DatumaroBase, JsonReader


class DatumaroBinaryBase(DatumaroBase):
    """"""

    def __init__(
        self,
        path: str,
        *,
        encryption_key: Optional[bytes] = None,
        num_workers: int = 0,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        """
        Parameters
        ----------
        path
            Directory path to import DatumaroBinary format dataset
        encryption_key
            If the dataset is encrypted, it (secret key) is needed to import the dataset.
        num_workers
            The number of multi-processing workers for import. If num_workers = 0, do not use multiprocessing.
        """
        self._fp: Optional[BufferedReader] = None
        self._crypter = Crypter(encryption_key) if encryption_key is not None else NULL_CRYPTER
        self._media_encryption = False
        self._num_workers = num_workers
        super().__init__(path, subset=subset, ctx=ctx)

    def _get_dm_format_version(self, path: str) -> str:
        with open(path, "rb") as fp:
            self._fp = fp
            self._check_signature()
            dm_format_version = self._read_version()
        return dm_format_version

    def _load_impl(self, path: str) -> None:
        """Actual implementation of loading Datumaro binary format."""
        try:
            with open(path, "rb") as fp:
                self._fp = fp
                self._check_signature()
                self._read_version()
                self._check_encryption_field()
                self._read_info()
                self._read_categories()
                self._read_media_type()
                self._read_items()
        finally:
            self._fp = None

    def _check_signature(self):
        signature = self._fp.read(DatumaroBinaryPath.SIGNATURE_LEN).decode()
        DatumaroBinaryPath.check_signature(signature)

    def _check_encryption_field(self):
        len_byte = self._fp.read(4)
        _bytes = self._fp.read(struct.unpack("I", len_byte)[0])

        if not self._crypter.handshake(_bytes):
            raise DatasetImportError("Encryption key handshake fails. You give a wrong key.")

    def _read_header(self, use_crypter: bool = True):
        len_byte = self._fp.read(4)
        _bytes = self._fp.read(struct.unpack("I", len_byte)[0])
        if use_crypter:
            _bytes = self._crypter.decrypt(_bytes)
        header, _ = DictMapper.backward(_bytes)
        return header

    def _read_version(self) -> Dict[str, Any]:
        version_header = self._read_header(use_crypter=False)
        self._media_encryption = version_header["media_encryption"]
        return version_header["dm_format_version"]

    def _read_info(self):
        self._infos = self._read_header()

    def _read_categories(self):
        categories = self._read_header()
        self._categories = JsonReader._load_categories({"categories": categories})

    def _read_media_type(self):
        media_type = self._read_header()["media_type"]
        if media_type == MediaType.IMAGE:
            self._media_type = Image
        elif media_type == MediaType.POINT_CLOUD:
            self._media_type = PointCloud
        elif media_type == MediaType.VIDEO:
            self._media_type = Video
        elif media_type == MediaType.VIDEO_FRAME:
            self._media_type = VideoFrame
        elif media_type == MediaType.MEDIA_ELEMENT:
            self._media_type = MediaElement
        else:
            raise NotImplementedError(f"media_type={media_type} is currently not supported.")

    def _read_items(self) -> None:
        (n_blob_sizes_bytes,) = struct.unpack("<I", self._fp.read(4))
        blob_sizes_bytes = self._crypter.decrypt(self._fp.read(n_blob_sizes_bytes))
        blob_sizes, _ = IntListMapper.backward(blob_sizes_bytes, 0)

        media_path_prefix = {
            MediaType.IMAGE: osp.join(self._images_dir, self._subset),
            MediaType.POINT_CLOUD: osp.join(self._pcd_dir, self._subset),
            MediaType.VIDEO: osp.join(self._video_dir, self._subset),
            MediaType.VIDEO_FRAME: osp.join(self._video_dir, self._subset),
        }

        if self._num_workers > 0:
            self._items = self._read_items_mp(blob_sizes, media_path_prefix)
        else:
            self._items = self._read_items_sp(blob_sizes, media_path_prefix)

        for item in self._items:
            if item.media is not None and self._media_encryption:
                item.media.set_crypter(self._crypter)

            for ann in item.annotations:
                self._ann_types.add(ann.type)

    def _read_items_mp(
        self, blob_sizes: List[int], media_path_prefix: Dict[MediaType, str]
    ) -> List[DatasetItem]:
        async_results: List[AsyncResult] = []

        with Pool(processes=self._num_workers) as pool:
            for blob_size in blob_sizes:
                blob_bytes = self._fp.read(blob_size)
                async_results += [
                    pool.apply_async(
                        self._read_blob,
                        (
                            blob_bytes,
                            self._crypter,
                            media_path_prefix,
                        ),
                    )
                ]

            return [
                item
                for async_result in async_results
                for item in async_result.get(timeout=DatumaroBinaryPath.MP_TIMEOUT)
            ]

    def _read_items_sp(
        self, blob_sizes: List[int], media_path_prefix: Dict[MediaType, str]
    ) -> List[DatasetItem]:
        items_list = [
            self._read_blob(self._fp.read(blob_size), self._crypter, media_path_prefix)
            for blob_size in blob_sizes
        ]

        return [item for items in items_list for item in items]

    @staticmethod
    def _read_blob(
        blob_bytes: bytes, crypter: Crypter, media_path_prefix: Dict[MediaType, str]
    ) -> List[DatasetItem]:
        items = []
        offset = 0

        # Decrypt bytes first
        blob_bytes = crypter.decrypt(blob_bytes)

        # Extract items
        while offset < len(blob_bytes):
            item, offset = DatasetItemMapper.backward(blob_bytes, offset, media_path_prefix)
            items.append(item)

        assert offset == len(blob_bytes)

        return items

    @property
    def is_stream(self) -> bool:
        return False

    def infos(self):
        return self._infos

    def categories(self):
        return self._categories

    def media_type(self):
        return self._media_type

    def ann_types(self):
        return self._ann_types

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> DatasetItem:
        yield from self._items



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from glob import glob
from typing import Optional

from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import ImportContext
from datumaro.plugins.data_formats.coco.base import _CocoBase
from datumaro.plugins.data_formats.coco.format import CocoImporterType, CocoTask
from datumaro.plugins.data_formats.coco.importer import CocoImporter


class MmdetCocoImporter(CocoImporter):
    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> FormatDetectionConfidence:
        ann_paths = context.require_files("annotations/instances_*.json")

        for ann_path in ann_paths:
            subset_name = cls._get_subset_name(ann_path)

            with context.require_any():
                with context.alternative():
                    image_files = osp.join(subset_name, "*.jpg")
                    context.require_file(f"{image_files}")

                    return FormatDetectionConfidence.MEDIUM

    def __call__(self, path, stream: bool = False, **extra_params):
        subset_paths = glob(osp.join(path, "**", "instances_*.json"), recursive=True)

        sources = []
        for subset_path in subset_paths:
            options = dict(extra_params)
            options["subset"] = self._get_subset_name(subset_path)

            if stream:
                options["stream"] = True

            sources.append({"url": subset_path, "format": "mmdet_coco", "options": options})

        return sources

    @classmethod
    def _get_subset_name(cls, subset_path: str):
        parts = osp.splitext(osp.basename(subset_path))[0].split("instances_", maxsplit=1)
        subset_name = parts[1] if len(parts) == 2 else DEFAULT_SUBSET_NAME

        return subset_name


class MmdetCocoBase(_CocoBase):
    """
    Parses Roboflow COCO annotations written in the following format:
    https://cocodataset.org/#format-data
    """

    def __init__(
        self,
        path,
        *,
        subset: Optional[str] = None,
        stream: bool = False,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(
            path,
            task=CocoTask.instances,
            coco_importer_type=CocoImporterType.mmdet,
            subset=subset,
            stream=stream,
            ctx=ctx,
        )



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse

from datumaro.util.scope import scope_add, scoped

from ....util.project import load_project


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(description="Prints project history.")

    parser.add_argument(
        "-n",
        "--max-count",
        default=10,
        type=int,
        help="Count of last commits to print (default: %(default)s)",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=log_command)

    return parser


def get_sensitive_args():
    return {
        log_command: [
            "project_dir",
        ],
    }


@scoped
def log_command(args):
    project = scope_add(load_project(args.project_dir))

    revisions = project.history(args.max_count)
    if revisions:
        for rev, message in revisions:
            print("%s %s" % (rev, message))
    else:
        print("(Project history is empty)")

    return 0



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from itertools import groupby

from datumaro.components.annotation import AnnotationType, LabelCategories

from .representation import import_predictions

try:
    from openvino.tools.accuracy_checker.adapters import create_adapter
    from openvino.tools.accuracy_checker.data_readers import DataRepresentation
    from openvino.tools.accuracy_checker.launcher import InputFeeder, create_launcher
    from openvino.tools.accuracy_checker.postprocessor import PostprocessingExecutor
    from openvino.tools.accuracy_checker.preprocessor import PreprocessingExecutor
    from openvino.tools.accuracy_checker.utils import extract_image_representations

    class _FakeDataset:
        def __init__(self, metadata=None):
            self.metadata = metadata or {}

    class GenericAcLauncher:
        @staticmethod
        def from_config(config):
            launcher_config = config["launcher"]
            launcher = create_launcher(launcher_config)

            dataset = _FakeDataset()
            adapter_config = config.get("adapter") or launcher_config.get("adapter")
            label_config = (
                adapter_config.get("labels") if isinstance(adapter_config, dict) else None
            )
            if label_config:
                assert isinstance(label_config, (list, dict))
                if isinstance(label_config, list):
                    label_config = dict(enumerate(label_config))

                dataset.metadata = {
                    "label_map": {int(key): label for key, label in label_config.items()}
                }
            adapter = create_adapter(adapter_config, launcher, dataset)

            preproc_config = config.get("preprocessing")
            preproc = None
            if preproc_config:
                preproc = PreprocessingExecutor(
                    preproc_config,
                    dataset_meta=dataset.metadata,
                    input_shapes=launcher.inputs_info_for_meta(),
                )

            postproc_config = config.get("postprocessing")
            postproc = None
            if postproc_config:
                postproc = PostprocessingExecutor(
                    postproc_config,
                    dataset_meta=dataset.metadata,
                )

            return __class__(launcher, adapter=adapter, preproc=preproc, postproc=postproc)

        def __init__(self, launcher, adapter=None, preproc=None, postproc=None, input_feeder=None):
            self._launcher = launcher
            self._input_feeder = input_feeder or InputFeeder(
                launcher.config.get("inputs", []),
                launcher.inputs,
                launcher.fit_to_input,
                launcher.default_layout,
            )
            self._adapter = adapter
            self._preproc = preproc
            self._postproc = postproc

            self._categories = self._init_categories()

        def launch_raw(self, inputs):
            ids = range(len(inputs))
            inputs = [DataRepresentation(inp, identifier=id) for id, inp in zip(ids, inputs)]
            _, batch_meta = extract_image_representations(inputs)

            if self._preproc:
                inputs = self._preproc.process(inputs)

            inputs = self._input_feeder.fill_inputs(inputs)
            outputs = self._launcher.predict(inputs, batch_meta)

            if self._adapter:
                outputs = self._adapter.process(outputs, ids, batch_meta)

            if self._postproc:
                outputs = self._postproc.process(outputs)

            return outputs

        def launch(self, inputs):
            outputs = self.launch_raw(inputs)
            return [import_predictions(g) for _, g in groupby(outputs, key=lambda o: o.identifier)]

        def categories(self):
            return self._categories

        def _init_categories(self):
            if self._adapter is None or self._adapter.label_map is None:
                return None

            label_map = sorted(self._adapter.label_map.items(), key=lambda e: e[0])

            label_cat = LabelCategories()
            for _, label in label_map:
                label_cat.add(label)

            return {AnnotationType.label: label_cat}

except ImportError:

    class GenericAcLauncher:
        def __init__(self):
            raise ImportError(
                "Accuracy Checker was deprecated from OpenVINO 2024.0. Please use lower OpenVINO version."
            )
